
/* MatchKeys.cs
 *
 * Keypoint file correlation functionality.
 *
 * (C) Copyright 2004 -- Sebastian Nowozin (nowozin@cs.tu-berlin.de)
 *
 * "This software is provided for non-commercial use only. The University of
 * British Columbia has applied for a patent on the SIFT algorithm in the
 * United States. Commercial applications of this software may require a
 * license from the University of British Columbia."
 * For more information, see the LICENSE file supplied with the distribution.
 */

using System;
using System.IO;
using System.Collections;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;


public class MatchKeys
{
	public ArrayList FindMatchesBBF (KeypointN[] keys1, KeypointN[] keys2)
	{
		// TODO: swap so smaller list is searched.

		ArrayList al = new ArrayList ();
		foreach (KeypointN kp in keys2)
			al.Add (kp);
		KDTree kd = KDTree.CreateKDTree (al);

		ArrayList matches = new ArrayList ();
		foreach (KeypointN kp in keys1) {
			ArrayList kpNNList = kd.NearestNeighbourListBBF (kp, 2, 40);

			if (kpNNList.Count < 2)
				throw (new Exception ("BUG: less than two neighbours!"));

			KDTree.BestEntry be1 = (KDTree.BestEntry) kpNNList[0];
			KDTree.BestEntry be2 = (KDTree.BestEntry) kpNNList[1];

			if ((be1.Distance / be2.Distance) > 0.6)
				continue;

			KeypointN kpN = (KeypointN) be1.Neighbour;

			matches.Add (new Match (kp, kpN, be1.Distance, be2.Distance));

			/*
			Console.WriteLine ("({0},{1}) ({2},{3}) {4}, 2nd: {5}", (int)(kp.X + 0.5),
				(int)(kp.Y + 0.5), (int)(kpN.X + 0.5),
				(int)(kpN.Y + 0.5), be1.Distance, be2.Distance);
			*/
		}

		return (matches);
	}

	/*
	public ArrayList FindMatches (ArrayList keys1, ArrayList keys2)
	{
		ArrayList matches = new ArrayList ();

		//KDTree kd = KDTree.CreateKDTree (keys2);
		foreach (Keypoint kp in keys1) {
			double distNearest = Double.PositiveInfinity;
			int nearest = -1;
			double dist2Nearest = Double.PositiveInfinity;
			int nearest2 = -1;

			for (int kn = 0 ; kn < keys2.Count ; ++kn) {
				Keypoint kp2 = (Keypoint) keys2[kn];

				double dist = Math.Sqrt (KDTree.DistanceSq (kp, kp2));
				if (dist < distNearest) {
					nearest2 = nearest;
					dist2Nearest = distNearest;

					nearest = kn;
					distNearest = dist;
				}
			}

			if (nearest == -1 || nearest2 == -1)
				continue;

			if ((distNearest / dist2Nearest) > 0.6)
				continue;

			matches.Add (new Match (kp, (Keypoint) keys2[nearest]));

			Console.WriteLine ("({0},{1}) ({2},{3}) {4}", (int)(kp.X + 0.5),
				(int)(kp.Y + 0.5), (int)(((Keypoint) keys2[nearest]).X + 0.5),
				(int)(((Keypoint) keys2[nearest]).Y + 0.5), distNearest);

		}

		return (matches);
	}
	*/

	public static ArrayList FilterJoins (ArrayList matches)
	{
		Hashtable ht = new Hashtable ();

		// Count the references to each keypoint
		foreach (Match m in matches) {
			int lI = (ht[m.Kp1] == null) ? 0 : (int) ht[m.Kp1];
			ht[m.Kp1] = lI + 1;
			int rI = (ht[m.Kp2] == null) ? 0 : (int) ht[m.Kp2];
			ht[m.Kp2] = rI + 1;
		}

		ArrayList survivors = new ArrayList ();
		int removed = 0;
		foreach (Match m in matches) {
			//Console.WriteLine ("match: {0}, {1}", (int) ht[m.Kp1], (int) ht[m.Kp2]);

			if (((int) ht[m.Kp1]) <= 1 && ((int) ht[m.Kp2]) <= 1)
				survivors.Add (m);
			else
				removed += 1;
		}
		Console.WriteLine ("{0} keypoint matches join-removed, {1} remaining",
			removed, survivors.Count);

		return (survivors);
	}

	public static void FilterNBest (ArrayList matches, int bestQ)
	{
		matches.Sort (new Match.MatchWeighter ());
		if (matches.Count > bestQ)
			matches.RemoveRange (bestQ, matches.Count - bestQ);
	}
}


public class Match
{
	KeypointN kp1;
	KeypointN kp2;
	public KeypointN Kp1 {
		get {
			return (kp1);
		}
	}
	public KeypointN Kp2 {
		get {
			return (kp2);
		}
	}

	double dist1;
	public double Dist1 {
		get {
			return (dist1);
		}
	}
	double dist2;
	public double Dist2 {
		get {
			return (dist2);
		}
	}

	private Match ()
	{
	}

	// dist1: distance between kp1/kp2,
	// dist2: distance between kp1 and kp3, where kp3 is the next closest
	//   match
	public Match (KeypointN kp1, KeypointN kp2, double dist1, double dist2)
	{
		this.kp1 = kp1;
		this.kp2 = kp2;
		this.dist1 = dist1;
		this.dist2 = dist2;
	}

	public class MatchWeighter : IComparer
	{
		private double distExp;
		private double quotExp;

		public MatchWeighter ()
			: this (1.0, 1.0)
		{
		}

		// The formula goes like this, with lowest weight being best matches:
		// w(kp) = kp.dist1^{distExp} *
		//     {\frac{1}{kp.dist2 - kp.dist1}}^{quotExp}
		//
		// This means, as both dist1 and dist2 are in [0.0 ; 1.0], that a high
		// distExp exponent (and distExp > quotExp) will make the absolute
		// distance for the best match more important. A high value for
		// quotExp will make the difference between the best and second best
		// match more important (as in "how many other candidates are likely
		// matches?").
		public MatchWeighter (double distExp, double quotExp)
		{
			this.distExp = distExp;
			this.quotExp = quotExp;
		}

		public int Compare (object obj1, object obj2)
		{
			Match m1 = (Match) obj1;
			Match m2 = (Match) obj2;

			double fit1 = OverallFitness (m1);
			double fit2 = OverallFitness (m2);
			if (fit1 < fit2)
				return (-1);
			else if (fit1 > fit2)
				return (1);

			return (0);
		}

		public double OverallFitness (Match m)
		{
			double fitness = Math.Pow (m.Dist1, distExp) *
				Math.Pow (1.0 / (m.Dist2 - m.Dist1), quotExp);

			return (fitness);
		}
	}
}


public class MultiMatch
{
	KeypointXMLList[] keysets;
	public KeypointXMLList[] Keysets {
		get {
			return (keysets);
		}
	}

	// Global k-d tree, containing all keys.
	KDTree globalKeyKD;
	// Global key list, containing Keypoint elements
	ArrayList globalKeys;

	// Global match list containing Match objects
	ArrayList globalMatches;

	// Partitioned matches
	MatchSet[,] matchSets;
	int imageCount;

	ArrayList filteredMatchSets;

	public void LoadKeysetsFromMemory (ArrayList memlist)
	{
		imageCount = memlist.Count;
		keysets = (KeypointXMLList[]) memlist.ToArray (typeof (KeypointXMLList));
	}

	public void LoadKeysets (string[] filenames)
	{
		imageCount = filenames.Length;

		keysets = new KeypointXMLList[imageCount];
		for (int n = 0 ; n < imageCount ; ++n) {
			KeypointXMLList keys = KeypointXMLReader.ReadComplete (filenames[n]);
			Console.WriteLine ("Loaded {0} keypoints from {1} for image {2} ({3}x{4})",
				keys.Arr.Length, filenames[n], keys.ImageFile, keys.XDim, keys.YDim);

			keysets[n] = keys;
		}
	}

	// Matches between two images
	public class MatchSet
	{
		string file1;
		public string File1 {
			get {
				return (file1);
			}
		}
		string file2;
		public string File2 {
			get {
				return (file2);
			}
		}

		ArrayList matches;
		public ArrayList Matches {
			get {
				return (matches);
			}
			set {
				matches = value;
			}
		}

		private MatchSet ()
		{
		}

		public MatchSet (string file1, string file2)
		{
			this.file1 = file1;
			this.file2 = file2;
			this.matches = new ArrayList ();
		}
	}

	// Returns a list of MatchSet objects
	public ArrayList LocateMatchSets (int minimumMatches, int bestMatches)
	{
		BuildGlobalKD ();
		BuildGlobalMatchList ();

		PartitionMatches ();

		filteredMatchSets = new ArrayList ();

		for (int n0 = 0 ; n0 < imageCount ; ++n0) {
			for (int n1 = n0 + 1 ; n1 < imageCount ; ++n1) {
				MatchSet ms = matchSets[n0, n1];
				if (ms == null)
					continue;

				int beforeFilterCount = ms.Matches.Count;

				Console.WriteLine ("Filtering... ({0}, {1})", ms.File1, ms.File2);
				ms.Matches = FilterMatches (ms.Matches, bestMatches);
				Console.WriteLine ("Filtered partition [{0},{1}] from {2} matches down to {3}",
					n0, n1, beforeFilterCount, ms.Matches.Count);

				// TODO: make threshhold configurable
				if (ms.Matches.Count >= minimumMatches)
					filteredMatchSets.Add (ms);
			}
		}

		return (filteredMatchSets);
	}

	private ArrayList FilterMatches (ArrayList matches, int bestMatches)
	{
		matches = MatchKeys.FilterJoins (matches);
		MatchKeys.FilterNBest (matches, bestMatches);

		Match.MatchWeighter mw = new Match.MatchWeighter ();

		foreach (Match m in matches) {
			Console.WriteLine ("    ({0},{1}) ({2},{3}) {4}, 2nd: {5}, weight {6}",
				(int)(m.Kp1.X + 0.5), (int)(m.Kp1.Y + 0.5),
				(int)(m.Kp2.X + 0.5), (int)(m.Kp2.Y + 0.5),
				m.Dist1, m.Dist2, mw.OverallFitness (m));
		}

		return (matches);
	}

	private void BuildGlobalKD ()
	{
		globalKeys = new ArrayList ();
		foreach (KeypointXMLList list in keysets)
			foreach (KeypointN kp in list.Arr)
				globalKeys.Add (kp);

		globalKeyKD = KDTree.CreateKDTree (globalKeys);
		Console.WriteLine ("Created global k-d tree containing {0} keypoints",
			globalKeys.Count);
	}

	// Partition the matches into image pair groups
	private void PartitionMatches ()
	{
		matchSets = new MatchSet[imageCount, imageCount];
		int createCount = 0;

		// Create all possible partition combinations, while pruning reverse
		// matches
		foreach (Match m in globalMatches) {
			int l0 = FindOrigin (m.Kp1);
			int l1 = FindOrigin (m.Kp2);

			bool reverseAlreadyIn = false;

			if (matchSets[l1, l0] != null) {
				MatchSet rev = matchSets[l1, l0];
				if (rev != null) {
					foreach (Match mr in rev.Matches) {
						if (mr.Kp1 == m.Kp2 && mr.Kp2 == m.Kp1) {
							reverseAlreadyIn = true;

							break;
						}
					}
				}
			}
			if (reverseAlreadyIn)
				continue;

			if (matchSets[l0, l1] == null) {
				createCount += 1;
				matchSets[l0, l1] = new MatchSet (keysets[l0].ImageFile,
					keysets[l1].ImageFile);
			}

			matchSets[l0, l1].Matches.Add (m);
		}

		Console.WriteLine ("Created {0} match partitions", createCount);
	}

	private int FindOrigin (KeypointN kp)
	{
		for (int ksn = 0 ; ksn < keysets.Length ; ++ksn) {
			KeypointXMLList list = keysets[ksn];

			int lIdx = Array.IndexOf (list.Arr, kp);
			if (lIdx >= 0)
				return (ksn);
		}

		throw (new Exception ("BUG: keypoint origin unknown"));
	}

	public delegate void MatchKeypointEventHandler (int index, int total);
	public event MatchKeypointEventHandler MatchKeypoint;

	private void BuildGlobalMatchList ()
	{
		globalMatches = new ArrayList ();

		int count = 0;
		foreach (KeypointN kp in globalKeys) {
			// Raise event in case an event handler has been registered.
			if (MatchKeypoint != null)
				MatchKeypoint (count, globalKeys.Count);

			if ((count % 25) == 0)
				Console.Write ("\r% {0}, {1}/{2}        ",
					(100 * ((double) count)) / ((double) globalKeys.Count),
					count, globalKeys.Count);
			count++;

			// There should be one exact hit (the keypoint itself, which we
			// ignore) and two real hits, so lets search for the three best
			// hits. But it could be the exact match is not found for, as we
			// use probabilistic bbf matching.
			ArrayList kpNNList = globalKeyKD.NearestNeighbourListBBF (kp, 3, 130);

			if (kpNNList.Count < 3)
				throw (new Exception ("BUG: less than three neighbours!"));

			KDTree.BestEntry be1 = (KDTree.BestEntry) kpNNList[0];
			KDTree.BestEntry be2 = (KDTree.BestEntry) kpNNList[1];

			// If be1 is the same (exact hit), shift one
			if (be1.Neighbour == kp) {
				be1 = be2;
				be2 = (KDTree.BestEntry) kpNNList[2];
			}
			if (be1.Neighbour == kp || be2.Neighbour == kp ||
				be1.Neighbour == be2.Neighbour)
				continue;
				//throw (new Exception ("BUG: wrong keypoints caught"));

			if ((be1.Distance / be2.Distance) > 0.6)
				continue;

			globalMatches.Add (new Match (kp, (KeypointN) be1.Neighbour,
				be1.Distance, be2.Distance));
		}

		Console.WriteLine ("\nGlobal match search yielded {0} matches",
			globalMatches.Count);
	}
}


